package eeds;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Stack;

import main.EDS;

import xmlgp.Experiment;
import xmlgp.Generation;
import xmlgp.Individual;
import xmlgp.Node;
import xmlgp.NodeType;
import xmlgp.Project;
import xmlgp.Representation;
import xmlgp.Treatment;
import xmlgp.Trial;
import xmlgp.XMLGP;

/**
 * This class simplifies GP results.
 * @author dharani
 *
 */
public final class SimplifyOutput {
		
	private String inputXmlFileName;
	private String outputXmlFileName;
	private String testPoints;
	private String operandFileName;
	private String operatorFileName;
	private String equivalanceNodeFileName;
	public static Stack<String> exprStack = new Stack<String>();
			
	//constructor
	SimplifyOutput(String inputXmlFileName, String outputXmlFileName, String testPoints, String operandFileName, String operatorFileName, String equivalanceNodeFileName)
	{
		this.inputXmlFileName = inputXmlFileName;
		this.outputXmlFileName = outputXmlFileName;
		this.testPoints = testPoints;
		this.operandFileName = operandFileName;
		this.operatorFileName = operatorFileName;
		this.equivalanceNodeFileName = equivalanceNodeFileName;		
	}
	
	//constructor
	SimplifyOutput(String testPoints, String operandFileName, String operatorFileName, String equivalanceNodeFileName)
	{		
		this.testPoints = testPoints;
		this.operandFileName = operandFileName;
		this.operatorFileName = operatorFileName;
		this.equivalanceNodeFileName = equivalanceNodeFileName;		
	}
	
	/**
	 * This method simplifies the GP results from input stream.
	 */
	private void simplifyGPResultFromStream() 
	{
		BufferedInputStream bis = null;
		DataInputStream dis = null;
		
		try
		{
			bis = new BufferedInputStream(System.in);
			dis = new DataInputStream(bis);	 
			deserialize(dis);
		}
		catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		try
    		{       			
    			dis.close();
    			bis.close();  			  			 	 
    		}
    		catch(IOException ioe)
    		{
    			ioe.printStackTrace();
    		}
    	}  
	}	
	
	/**
	 * This method simplifies the GP results from input XML file.
	 */
	private void simplifyGPResultFromFile() 
	{
		FileInputStream fis = null;
    	BufferedInputStream bis = null;
    	DataInputStream dis = null;
		
		try
		{
			fis = new FileInputStream(inputXmlFileName);
	        bis = new BufferedInputStream(fis);
	        dis = new DataInputStream(bis);	 
	        deserialize(dis);
	        
		}
		catch(FileNotFoundException fnfe)
    	{
    		fnfe.printStackTrace();
    	}
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		try
    		{       			
    			dis.close();
    			bis.close();
    			fis.close();   			 	 
    		}
    		catch(IOException ioe)
    		{
    			ioe.printStackTrace();
    		}
    	}  
	}
	
	/**
	 * Deserializes the input stream into an object(project)
	 * and constructs the simplified tree and replaces node in project.
	 * @param dataInputStream
	 */
    private void deserialize(DataInputStream dataInputStream) 
    {  
    	try
    	{	                   	
			Project project = XMLGP.deserializeFromStream(dataInputStream);				
        	formSimplifiedTree(project);
        	
        	if(null != outputXmlFileName)
        	{
        		XMLGP.serializeToFile(project, outputXmlFileName);   
        	}
        	else
        	{
        		XMLGP.serializeToStream(project);
        	}           
    	}    	
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
        
    }    
      
    /**
     * Simplifies the expression node and replaces the node in the input
     * @param project
     * @return
     */
    private void formSimplifiedTree(Project project)
    {   
       String expression = null;
       Node exprNode = null;    	
       
       try
       { 	      	
	    	for(Experiment experiment : project.getExperiments())
	    	{
	    		for(Treatment treatment : experiment.getTreatments())
		    	{
	    			for(Trial trial : treatment.getTrials())
			    	{
	    				for(Generation generation : trial.getGenerations())
				    	{
					    	for(Individual individual : generation.getIndividuals())
					    	{	   	
					    		for(Representation representation : individual.getRepresentations())
					    		{  					    		     
					    			exprNode = representation.getNode();					    			
					    						   				
									//tree to String					
									convertNodeToString(exprNode);
									
									if(!exprStack.empty())
									{
										expression = exprStack.pop();
										exprStack.clear();
										
										//System.out.println("Expression "+expression);
										exprNode = getSimplifiedExpression(expression);
										
					    				representation.setNode(exprNode);	
									}									
					    		}	    		
					    	}
				    	}
			    	}
		    	}
	    	}    	
       }
       catch(Exception e)
       {       	   
    	   e.printStackTrace();
       }      
    }
    
   /**
    * Converts the expression node to postfix string format
    * @param exprNode expression node
    */
    private void convertNodeToString(Node exprNode)
    {     	
		String lOperand = null;
		String rOperand = null;
		StringBuilder expr = null;
		String operator = null;
	
		if (null != exprNode)
		{
			ArrayList<Node> children = (ArrayList<Node>)exprNode.getChildren();
			Node cNode = null;			
				
	    	if (exprNode.getType() == NodeType.TERMINAL) 
	    	{	    		
	    		exprStack.push(exprNode.getValue());
	        }
	    	else
	    	{ 
	    		operator = exprNode.getValue();   		
	    		
	    		if(null != children)
	    		{	    		
	    			cNode = children.get(0);	    		
		    		convertNodeToString(cNode);		    		
			    	lOperand = exprStack.pop();		    		
		    		
		    		if(!"S".equals(operator))
		    		{
			    		cNode = children.get(1);			    		
		    			convertNodeToString(cNode);	    		
			    		rOperand = exprStack.pop();		    			
		    		}   		
	    		}
	    		
	    		expr = new StringBuilder();
	    		expr.append(lOperand);
	    		expr.append(" ");
	    		if(null != rOperand)
	    		{
	    			expr.append(rOperand);
	    			expr.append(" ");
	    		}
	    		expr.append(exprNode.getValue());
	    		
	    		exprStack.push(expr.toString());
	    	}   
		}
    }
       
    /**
     * Gets the simplified expression node for the given postfix expression string.
     * @param expression
     * @param operandFileName
     * @param operatorFileName
     * @param equivalanceNodeFileName
     * @return node
     */
    private Node getSimplifiedExpression(String expression)
    {
    	Node node = null;
		String result = null;
		
		try 
		{			
			//call Naoki's framework to simplify expression			
			EDS eds = new EDS();
			result = eds.simplifyGPResults(expression, testPoints, operandFileName, operatorFileName, equivalanceNodeFileName);			
			node = Util.constructNode(result.trim());			
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}	
		
		return node;
    }   
	
	/**
	 * The main method.
	 * @param args the command line arguments
	 */
	public static void main(String[] args) 
    {
    	SimplifyOutput eds = null;
    	int argLen = args.length;
		
		if (argLen == 6) 
		{
			eds = new SimplifyOutput(args[0], args[1], args[2], args[3], args[4], args[5]);
			eds.simplifyGPResultFromFile();  
		}
		else if (argLen == 4)
		{
			eds = new SimplifyOutput(args[0], args[1], args[2], args[3]);
			eds.simplifyGPResultFromStream();
		}
    }
}